//HavokManager.cpp
#include <Common/Base/hkBase.h>
#include <Common/Base/Container/String/hkStringBuf.h>
#include <Common/Base/Ext/hkBaseExt.h>
#include <Common/Base/Memory/System/hkMemorySystem.h>
#include <Common/Base/Memory/System/Util/hkMemoryInitUtil.h>
#include <Common/Base/Memory/Allocator/Malloc/hkMallocAllocator.h>
#include <Common/Serialize/Util/hkLoader.h>

#include <Physics/Dynamics/World/hkpWorld.h>
#include <Physics/Dynamics/Entity/hkpRigidBody.h>
#include <Physics/Dynamics/Entity/hkpRigidBodyCinfo.h>
#include <Physics/Utilities/Dynamics/Inertia/hkpInertiaTensorComputer.h>
#include <Physics/Collide/Shape/Convex/Box/hkpBoxShape.h>
#include <Physics/Collide/Dispatch/hkpAgentRegisterUtil.h>
#include <Physics/Collide/Filter/Group/hkpGroupFilter.h>
#include <Common/Serialize/Util/hkSerializeDeprecated.h>


//Ƽ 
#include <Common/Base/Thread/Job/ThreadPool/Cpu/hkCpuJobThreadPool.h>
#include <Common/Base/Thread/Job/ThreadPool/Spu/hkSpuJobThreadPool.h>
#include <Common/Base/Thread/JobQueue/hkJobQueue.h>


//  
#include <Common/Visualize/hkVisualDebugger.h>
#include <Physics/Utilities/VisualDebugger/hkpPhysicsContext.h>		

#include <assert.h>
#include "HavokManager.h"


//havok_config.cpp
#pragma comment(lib, "hkBase.lib")
#pragma comment(lib, "hkSerialize.lib")
#pragma comment(lib, "hkSceneData.lib")
#pragma comment(lib, "hkInternal.lib")
#pragma comment(lib, "hkGeometryUtilities.lib")
#pragma comment(lib, "hkVisualize.lib")
#pragma comment(lib, "hkCompat.lib")
#pragma comment(lib, "hkpCollide.lib")
#pragma comment(lib, "hkpconstraintsolver.lib")
#pragma comment(lib, "hkpDynamics.lib")
#pragma comment(lib, "hkpinternal.lib")
#pragma comment(lib, "hkputilities.lib")
#pragma comment(lib, "hkpVehicle.lib")


// [id=keycode]
#include <Common/Base/keycode.cxx>

// [id=productfeatures]
// We're using only physics - we undef products even if the keycode is present so
// that we don't get the usual initialization for these products.
#undef HK_FEATURE_PRODUCT_AI
#undef HK_FEATURE_PRODUCT_ANIMATION
#undef HK_FEATURE_PRODUCT_CLOTH
#undef HK_FEATURE_PRODUCT_DESTRUCTION
#undef HK_FEATURE_PRODUCT_BEHAVIOR
#define HK_FEATURE_REFLECTION_PHYSICS
#define HK_CLASSES_FILE <Common/Serialize/Classlist/hkClasses.h>

// Also we're not using any of these.
//#define HK_EXCLUDE_FEATURE_SerializeDeprecatedPre700
//#define HK_EXCLUDE_FEATURE_RegisterVersionPatches
	//#define HK_EXCLUDE_FEATURE_RegisterReflectedClasses
//#define HK_EXCLUDE_FEATURE_MemoryTracker
//#define HK_EXCLUDE_FEATURE_hkpAccurateInertiaTensorComputer
//#define HK_EXCLUDE_FEATURE_CompoundShape
//#define HK_EXCLUDE_FEATURE_hkpAabbTreeWorldManager
//#define HK_EXCLUDE_FEATURE_hkpContinuousSimulation
//#define HK_EXCLUDE_FEATURE_hkpKdTreeWorldManager


#include <Common/Base/Config/hkProductFeatures.cxx>

#include <stdio.h>


HavokManager* HavokManager::m_pInstance = 0L;

void ErrorReportFunction(const char* msg, void* errorReportObject) 
{
	char str[300];
	sprintf_s( str, 300, "%s\n", msg );
	OutputDebugString( str );
}
//--------------------------------------------------------------------------------
bool HavokManager::CreateInstance( bool bMultithread )
{
	if( m_pInstance )
		return false;

	m_pInstance = new HavokManager();
	bool bRes = m_pInstance->Init( bMultithread );
	return bRes;
}
//--------------------------------------------------------------------------------
void HavokManager::ReleaseInstance()
{
	if( m_pInstance == 0L )
		return;
	
	m_pInstance->UnInit();
	delete m_pInstance;
	m_pInstance = 0L;
}
//--------------------------------------------------------------------------------
HavokManager::HavokManager()
{
	m_bMultiThread = m_bActiveVisualDebugger = false;
	m_threadPool = NULL;
	m_jobQueue  = NULL;
	m_world = NULL;
	m_vdb = NULL;
	m_context = NULL;
	m_stopWatch = NULL;
	m_fLastTime = 0;
}
//--------------------------------------------------------------------------------
HavokManager::~HavokManager()
{
}
//--------------------------------------------------------------------------------
bool HavokManager::Init( bool bMultithread )
{
	m_bMultiThread = bMultithread;

	hkMemoryRouter* memoryRouter = 
		hkMemoryInitUtil::initDefault(hkMallocAllocator::m_defaultMallocAllocator, hkMemorySystem::FrameInfo(500000));
	hkBaseSystem::init( memoryRouter, ErrorReportFunction );

	if( bMultithread )
	{
		//get hardware info and initialize total number of threads used
		int totalNumThreadsUsed;
		hkHardwareInfo hwInfo;
		hkGetHardwareInfo(hwInfo);
		totalNumThreadsUsed = hwInfo.m_numThreads;

		//initialize thread pool (totalNumThreadsUsed- 1, one thread used for physics world simulation)
		hkCpuJobThreadPoolCinfo threadPoolCinfo;
		threadPoolCinfo.m_numThreads = totalNumThreadsUsed - 1;

		//enable timers collection, allocating 200 Kb per thread
		threadPoolCinfo.m_timerBufferPerThreadAllocation = 200000;
		this->m_threadPool = new hkCpuJobThreadPool( threadPoolCinfo );

		//create a job queue. This job queue will be used by all Havok modules to run multithreaded work.
		//In following tutorial it is used only for physics simulation
		hkJobQueueCinfo jobQueueCinfo;
		jobQueueCinfo.m_jobQueueHwSetup.m_numCpuThreads = totalNumThreadsUsed;
		this->m_jobQueue = new hkJobQueue(jobQueueCinfo);
	}

	//Enable monitors for the thread.
	//Monitors have been enabled above for thread pool threads 
	hkMonitorStream::getInstance().resize(200000);

	//initialize physics world - create world info
	hkpWorldCinfo info;

	if( bMultithread )
	{
		//set simualation type to multithreaded
		info.m_simulationType = hkpWorldCinfo::SIMULATION_TYPE_MULTITHREADED;
	}
	else
		info.m_simulationType = hkpWorldCinfo::SIMULATION_TYPE_CONTINUOUS;

	//objects that fall "out of the world" will be automatically removed
	info.m_broadPhaseBorderBehaviour = hkpWorldCinfo::BROADPHASE_BORDER_REMOVE_ENTITY;
	
	//standard gravity settings, collision tolerance and world size 
	info.m_gravity.set(0,-12.8f,0);
	info.m_collisionTolerance = 0.1f;; 
	info.setBroadPhaseWorldSize( 10000.0f );
	info.setupSolverInfo( hkpWorldCinfo::SOLVER_TYPE_8ITERS_HARD );
	
	//initialize world with created info
	this->m_world = new hkpWorld(info);
	this->m_pLoader = new hkLoader;

	//When the simulation type is SIMULATION_TYPE_MULTITHREADED, in the debug build, the sdk performs checks
	//to make sure only one thread is modifying the world at once to prevent multithreaded bugs. Each thread
	//must call markForRead / markForWrite before it modifies the world to enable these checks.
	this->m_world->markForWrite();

	//Register all collision agents
	//It's important to register collision agents before adding any entities to the world.
	hkpAgentRegisterUtil::registerAllAgents(m_world->getCollisionDispatcher());
	
	//register physics world with job queue
	if( m_jobQueue )
		this->m_world->registerWithJobQueue(m_jobQueue);

	// Replace filter
	hkpGroupFilter* groupFilter = new hkpGroupFilter();
	groupFilter->disableCollisionsBetween(LAYER_ACTOR, LAYER_ACTOR);
	m_world->setCollisionFilter( groupFilter, true);
	groupFilter->removeReference();

	//world no loger modified - unmark it
	this->m_world->unmarkForWrite();

	hkSerializeDeprecated::initDeprecated();

	m_stopWatch = new hkStopwatch( "HavokManager" );
	m_stopWatch->start();
	return true;
}
//--------------------------------------------------------------------------------
void HavokManager::UnInit()
{
	//clean up physics world
	this->m_world->markForWrite();
	this->m_world->removeReference();
	if(  this->m_pLoader )
	{
		delete this->m_pLoader;
		m_pLoader = NULL;
	}

	//clean up visual debugger(if used)
	if(this->m_bActiveVisualDebugger == true)
	{
		this->m_vdb->removeReference();
		this->m_context->removeReference();
	}

	//delete job queue, thread poll, deallocate memory
	if( m_jobQueue ) 
		delete m_jobQueue;
	if( m_threadPool )
		m_threadPool->removeReference();

	if( m_stopWatch )
		delete m_stopWatch;

	//quit base system
	hkBaseSystem::quit();

	hkMemoryInitUtil::quit();
}
//--------------------------------------------------------------------------------
void HavokManager::ActiveVisualDebugger()
{
	if( m_bActiveVisualDebugger == false )
	{
		// Create context for Visual Debugger
		m_context = new hkpPhysicsContext();

		//Reigster all the physics viewers
		hkpPhysicsContext::registerAllPhysicsProcesses();

		//Add physics world to context, so viewers can see it
		m_world->markForWrite();
		m_context->addWorld(m_world);
		m_world->unmarkForWrite();

		hkArray<hkProcessContext*> contexts;
		contexts.pushBack(m_context);

		//Create VDB instance
		m_vdb = new hkVisualDebugger(contexts);
		m_vdb->serve();

		m_bActiveVisualDebugger = true;
	}	
}
//--------------------------------------------------------------------------------
void HavokManager::Update()
{
	hkReal	timeStep = 1.f / 60.f;
	hkReal	fCurTime = m_stopWatch->getElapsedSeconds();
	int		nMaxLoop = 5;

	for( int n = 0; fCurTime - m_fLastTime >= timeStep; ++n )
	{
		StepSimulation( timeStep );
		StepVisualDebugger( timeStep );
		m_fLastTime += timeStep;
		if( n > nMaxLoop )
			break;
	}
}
//--------------------------------------------------------------------------------
void HavokManager::StepSimulation( float dt )
{
	if( m_bMultiThread )
	{
		//step multithreaded simulation using this thread and all threads in the thread pool
		m_world->stepMultithreaded(m_jobQueue, m_threadPool, dt);

		hkMonitorStream::getInstance().reset();
		m_threadPool->clearTimerData();	
	}
	else
		m_world->stepDeltaTime( dt );	
}
//--------------------------------------------------------------------------------
void HavokManager::StepVisualDebugger( float dt )
{
	if( m_bActiveVisualDebugger )
	{
		//synchronize the timer data and step Visual Debugger
		m_context->syncTimers( m_threadPool );
		m_vdb->step();
	}	
}
//--------------------------------------------------------------------------------
hkLoader*	HavokManager::GetLoader()
{
	return m_pLoader;
}


